#define WIN32_LEAN_AND_MEAN
#include <Windows.h>

#include <cstdio>
#include <cstdlib>
#include <cwchar>


template <typename FnDeleter>
class Deleter
{
private:
    FnDeleter m_deleter;

public:
    explicit Deleter(FnDeleter deleter) noexcept
        : m_deleter(deleter)
    {
    }

    ~Deleter() noexcept
    {
        m_deleter();
    }
};

unsigned int inject(unsigned int pid, const wchar_t* libPath) noexcept
{
    if (!libPath)
    {
        return ERROR_INVALID_PARAMETER;
    }

    const auto pathLength = wcslen(libPath);
    if (!pathLength)
    {
        return ERROR_INVALID_PARAMETER;
    }

    const HANDLE hProcess = OpenProcess(PROCESS_VM_OPERATION | PROCESS_VM_WRITE | PROCESS_CREATE_THREAD, false, pid);
    if (!hProcess)
    {
        return GetLastError();
    }

    const Deleter handleCloser([hProcess]()
    {
        CloseHandle(hProcess);
    });

    const auto pathSizeInBytes = pathLength * sizeof(wchar_t) + sizeof(L'\0');
    void* const pathBuffer = VirtualAllocEx(hProcess, nullptr, pathSizeInBytes, MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE);
    if (!pathBuffer)
    {
        return GetLastError();
    }

    const Deleter bufferDeleter([hProcess, pathBuffer]()
    {
        VirtualFreeEx(hProcess, pathBuffer, 0, MEM_RELEASE);
    });

    SIZE_T written = 0;
    const bool writeStatus = !!WriteProcessMemory(hProcess, pathBuffer, libPath, pathSizeInBytes, &written);
    if (!writeStatus)
    {
        return GetLastError();
    }

    const HMODULE hKernel32 = GetModuleHandleW(L"kernel32.dll");
    if (!hKernel32)
    {
        return GetLastError();
    }

    const void* const loadLibrary = GetProcAddress(hKernel32, "LoadLibraryW");
    if (!loadLibrary)
    {
        return GetLastError();
    }

    const HANDLE hThread = CreateRemoteThread(hProcess, nullptr, 0, reinterpret_cast<LPTHREAD_START_ROUTINE>(loadLibrary), pathBuffer, 0, nullptr);
    if (!hThread)
    {
        return GetLastError();
    }

    const Deleter threadCloser([hThread]()
    {
        CloseHandle(hThread);
    });

    WaitForSingleObject(hThread, INFINITE);

    return ERROR_SUCCESS;
}

int wmain(int argc, wchar_t* argv[])
{
    if (argc != 3)
    {
        printf("Usage:\n");
        printf("%ws \"N:\\Folder\\Lib.dll\" 1234\n", argv[0]);
        printf("  Where '1234' is the PID of the target process.\n");
        return 0;
    }

    const wchar_t* const libPath = argv[1];
    const unsigned int pid = static_cast<unsigned int>(std::wcstoull(argv[2], nullptr, 10));

    const auto status = inject(pid, libPath);
    printf("Status: 0x%X\n", status);

    return (status == ERROR_SUCCESS) ? 0 : -1;
}
